-------------------------------------------------
-- World View 
-------------------------------------------------

print("--- LightTouch mod ---");

include( "FLuaVector" );
include( "GlobalProperties" );
include( "LightTouchDiploCornerAddin" );

-- determine if Gods and Kings DLC is enabled
local g_GNK = false;
for row in GameInfo.DownloadableContent() do 
	if(row.PackageID == "0E3751A1F8404E1B9706519BF484E59D") then
		print("checking DLC: Gods and Kings DLC is enabled");
		g_GNK = true;
		break;
	end
end

-- define variables
local g_pathFromSelectedUnitToMouse = nil;
local turn1Color = Vector4( 0, 1, 0, 0.25 );
local turn2Color = Vector4( 1, 1, 0, 0.25 );
local turn3PlusColor = Vector4( 0.25, 0, 1, 0.25 );
local maxRangeColor = Vector4( 1, 1, 1, 0.25 );
local rButtonDown = false;

local pathBorderStyle = "MovementRangeBorder";
local attackPathBorderStyle = "AMRBorder"; -- attack move

local g_LTO_MissionOnStart_set_next_turn = nil;
local endTurnExecutedOnTurn = nil;

--LT MOD registries. If not local, it's for debugging purposes
local aMissionStack={};
local blackList={};

local highlightedTilesList={};		 
local tempDataTable={};

function dbprint(...)
	if(g_DEBUG) then
		print(...);
	end
end


local mapMaxX, mapMaxY = Map.GetGridSize();
local mapMaxXX= mapMaxX; --DEBUG
mapMaxX = 100;

-- Quicksort sorting algorithm
function Quicksort(array, left, right, sortindex)		--TODO: Quicksort could be made potentially quicker if insertion sort is used for small subarrays
	if(left==nil or right==nil) then
		left = 1;
		right = #array;
	end
	if(sortindex==nil) then
		sortindex = {};
		for i = 1, right do
			sortindex[i] = i;
		end
	end

	local pivotIndex = math.floor((left+right)/2);	--we will use the middle-most element of array as pivot.
	local newPivotIndex;

	if(right > left) then

		array, newPivotIndex, sortindex = QuicksortPartition(array, left, right, pivotIndex, sortindex);
		array, sortindex = Quicksort(array, left, newPivotIndex - 1, sortindex);
	else
		return array, sortindex;
	end

	return Quicksort(array, newPivotIndex + 1, right, sortindex);	--use tail call to sort right partition
end

function QuicksortPartition(array, left, right, pivotIndex, sortindex)
	local pivotval = array[pivotIndex];
	local pivotsi = sortindex[pivotIndex];
	array[pivotIndex] = array[right];
	sortindex[pivotIndex] = sortindex[right];
	array[right] = pivotval;
	sortindex[right] = pivotsi;

	local storeIndex = left;
	for i = left, right-1, 1 do
		if(array[i] < pivotval) then
			local storeval = array[storeIndex];
			array[storeIndex] = array[i];
			array[i] = storeval;

			local storesi = sortindex[storeIndex];
			sortindex[storeIndex] = sortindex[i];
			sortindex[i] = storesi;

			storeIndex = storeIndex + 1;
		end
	end

	array[right] = array[storeIndex];
	array[storeIndex] = pivotval;
	sortindex[right] = sortindex[storeIndex];
	sortindex[storeIndex] = pivotsi;

    return array, storeIndex, sortindex;
end


function SetInflate(set, radius)	-- boundary checked
	local newset={};
	for k, v in pairs(set) do
		newset[k] = {["x"] = v.x, ["y"] = v.y};
		newset = MarkSetX(newset, k, radius);
	end
	for k, v in pairs(newset) do
		if(v.x < 0 or v.x >= mapMaxXX or v.y < 0 or v.y >= mapMaxY) then
			newset[k] = nil;
		end
	end
	return newset;
end

function MarkSetX(set, k, radius)	-- NO boundary check
	local newk;
	local x = set[k].x;
	local y = set[k].y;
	local oddCorr = (set[k].y % 2 == 0) and -1 or 1;	--correction for odd rows displacement
	local leftCorr = math.min(0, oddCorr);
	local rightCorr = math.max(0, oddCorr);
	local startRange = math.floor((radius - 1) / 2);

	local widen = 0;
	for j = radius, 0, -1 do
		local oddOffset = j % 2;
		widen = widen + 1 - oddOffset;	--widen on even offsets

		local imin = -startRange - widen + leftCorr * oddOffset;	--correct on odd offsets
		local imax = startRange + widen + rightCorr * oddOffset;

		for i = imin, imax, 1 do
			newk = k + i + j * mapMaxX;
			local newx = x + i;
			if(mapWrapX) then newx = newx % mapMaxXX; end

			set[newk] = {["x"]=newx, ["y"]=y+j};
			if(j~=0) then
				newk = k + i - j * mapMaxX;
				set[newk] = {["x"]=newx, ["y"]=y-j};
			end
		end
	end
	return set;
end

function SetGenCircle(x, y, radi)
	local newset = {};
	newset[x + y * mapMaxX] = {["x"] = x, ["y"] = y};
	newset = SetInflate(newset, radi);
	return newset;
end

function SetSortByField(set, field)
	local prop={};
	local idlist={};
	local addPos = 1;
	for plotID, v in pairs(set) do
		prop[addPos] = v[field];
		idlist[addPos] = plotID;
		addPos = addPos + 1;
	end
	local propmap;
	prop, propmap = Quicksort(prop);

	local sortidlist = {};
	local sortidmap = {};
	for i, v in ipairs(propmap) do
		sortidlist[i] = idlist[v];
		sortidmap[idlist[v]] = i;
	end

	return prop, sortidlist, sortidmap;
end


function UpdatePathFromSelectedUnitToMouse()
	local interfaceMode = UI.GetInterfaceMode();
	--Events.ClearHexHighlightStyle(pathBorderStyle);
	--if (interfaceMode == InterfaceModeTypes.INTERFACEMODE_SELECTION and rButtonDown) or (interfaceMode == InterfaceModeTypes.INTERFACEMODE_MOVE_TO) then
		--for i, pathNode in ipairs(g_pathFromSelectedUnitToMouse) do
			--local hexID = ToHexFromGrid( Vector2( pathNode.x, pathNode.y) );
			--if pathNode.turn == 1 then
				--Events.SerialEventHexHighlight( hexID, true, turn1Color, pathBorderStyle );
			--elseif pathNode.turn == 2 then
				--Events.SerialEventHexHighlight( hexID, true, turn2Color, pathBorderStyle );
			--else
				--Events.SerialEventHexHighlight( hexID, true, turn3PlusColor, pathBorderStyle );
			--end
		--end
	
		-- this doesnt work because the size of the array is irrelevant
		--if ( #g_pathFromSelectedUnitToMouse > 0 ) then
			--Events.DisplayMovementIndicator( true );
		--else
			--Events.DisplayMovementIndicator( false );
		--end
	--end
	
	if (interfaceMode == InterfaceModeTypes.INTERFACEMODE_SELECTION and rButtonDown) or (interfaceMode == InterfaceModeTypes.INTERFACEMODE_MOVE_TO) then
		Events.DisplayMovementIndicator( true );
	else
		Events.DisplayMovementIndicator( false );
	end
	
end

function ShowMovementRangeIndicator()
	local unit = UI.GetHeadSelectedUnit();
	if not unit then
		return;
	end
	
	local iPlayerID = Game.GetActivePlayer();

	Events.ShowMovementRange( iPlayerID, unit:GetID() );
end

-- Add any interface modes that need special processing to this table
-- (look at InGame.lua for a bigger example)
local InterfaceModeMessageHandler = 
{
	[InterfaceModeTypes.INTERFACEMODE_DEBUG] = {},
	[InterfaceModeTypes.INTERFACEMODE_SELECTION] = {},
	[InterfaceModeTypes.INTERFACEMODE_PING] = {},
	[InterfaceModeTypes.INTERFACEMODE_MOVE_TO] = {},
	[InterfaceModeTypes.INTERFACEMODE_MOVE_TO_TYPE] = {},
	[InterfaceModeTypes.INTERFACEMODE_MOVE_TO_ALL] = {},
	[InterfaceModeTypes.INTERFACEMODE_ROUTE_TO] = {},
	[InterfaceModeTypes.INTERFACEMODE_AIRLIFT] = {},
	[InterfaceModeTypes.INTERFACEMODE_NUKE] = {},
	[InterfaceModeTypes.INTERFACEMODE_PARADROP] = {},
	[InterfaceModeTypes.INTERFACEMODE_ATTACK] = {},
	[InterfaceModeTypes.INTERFACEMODE_RANGE_ATTACK] = {},
	[InterfaceModeTypes.INTERFACEMODE_CITY_RANGE_ATTACK] = {},
	[InterfaceModeTypes.INTERFACEMODE_AIRSTRIKE] = {},
	[InterfaceModeTypes.INTERFACEMODE_AIR_SWEEP] = {},
	[InterfaceModeTypes.INTERFACEMODE_REBASE] = {},
	[InterfaceModeTypes.INTERFACEMODE_EMBARK] = {},
	[InterfaceModeTypes.INTERFACEMODE_DISEMBARK] = {},
	[InterfaceModeTypes.INTERFACEMODE_PLACE_UNIT] = {},
	[InterfaceModeTypes.INTERFACEMODE_GIFT_UNIT] = {}
}

local DefaultMessageHandler = {};


DefaultMessageHandler[KeyEvents.KeyDown] =
function( wParam, lParam )
	if ( wParam == Keys.VK_LEFT ) then
		Events.SerialEventCameraStopMovingRight();
		Events.SerialEventCameraStartMovingLeft();
		return true;
	elseif ( wParam == Keys.VK_RIGHT ) then
		Events.SerialEventCameraStopMovingLeft();
		Events.SerialEventCameraStartMovingRight();
		return true;
	elseif ( wParam == Keys.VK_UP ) then
		Events.SerialEventCameraStopMovingBack();
		Events.SerialEventCameraStartMovingForward();
		return true;
	elseif ( wParam == Keys.VK_DOWN ) then
		Events.SerialEventCameraStopMovingForward();
		Events.SerialEventCameraStartMovingBack();
		return true;
	elseif ( wParam == Keys.VK_NEXT or wParam == Keys.VK_OEM_MINUS ) then
		Events.SerialEventCameraOut( Vector2(0,0) );
		return true;
	elseif ( wParam == Keys.VK_PRIOR or wParam == Keys.VK_OEM_PLUS ) then
		Events.SerialEventCameraIn( Vector2(0,0) );
		return true;
	elseif ( wParam == Keys.VK_ESCAPE and InStrategicView() ) then
		ToggleStrategicView();
        return true;
	
	-- This ensures compatibility with Thal's VEM and UP mods
	elseif ( wParam == Keys.X and UIManager:GetControl() ) then
		--print("Control-X detected, activating the pin-board");
		LuaEvents.MapPins_Show()
		return true;
	elseif ( wParam == Keys.X and UIManager:GetShift() ) then
		--print("Shift-X detected, toggling pin visibility");
		LuaEvents.MapPins_Toggle()
		return true;
	end
end


DefaultMessageHandler[KeyEvents.KeyUp] =
function( wParam, lParam )
	if ( wParam == Keys.VK_LEFT ) then
		Events.SerialEventCameraStopMovingLeft();
        return true;
	elseif ( wParam == Keys.VK_RIGHT ) then
		Events.SerialEventCameraStopMovingRight();
        return true;
	elseif ( wParam == Keys.VK_UP ) then
		Events.SerialEventCameraStopMovingForward();
        return true;
	elseif ( wParam == Keys.VK_DOWN ) then
		Events.SerialEventCameraStopMovingBack();
        return true;
	end
	return false;
end


-- Emergency key up handler
function KeyUpHandler( wParam )
	if ( wParam == Keys.VK_LEFT ) then
		Events.SerialEventCameraStopMovingLeft();
	elseif ( wParam == Keys.VK_RIGHT ) then
		Events.SerialEventCameraStopMovingRight();
	elseif ( wParam == Keys.VK_UP ) then
		Events.SerialEventCameraStopMovingForward();
	elseif ( wParam == Keys.VK_DOWN ) then
		Events.SerialEventCameraStopMovingBack();
	end
end
Events.KeyUpEvent.Add( KeyUpHandler );

-- INTERFACEMODE_DEBUG


-- These updates to DEBUG TOOLS are from Thal's VEM and UP mods
g_UnitPlopper =
{
	UnitType = -1,
	Experience = 0,
	Embarked = false,
	
	Plop =
	function(plot)
		if (g_PlopperSettings.Player ~= -1 and g_UnitPlopper.UnitType ~= -1) then
			local player = Players[g_PlopperSettings.Player] or Players[Game.GetActivePlayer()];
			if (player ~= nil) then
				local unit = player:InitUnit(g_UnitPlopper.UnitType, plot:GetX(), plot:GetY());
				if (g_UnitPlopper.Embarked) then
					unit:Embark();
				end
				if g_UnitPlopper.Experience and g_UnitPlopper.Experience > 0 then
					unit:ChangeExperience(g_UnitPlopper.Experience);
					if g_UnitPlopper.Experience >= 10 then
						unit:SetPromotionReady(true);
					end
				end
			end
		end
	end,
	
	Deplop =
	function(plot)
		local unitCount = plot:GetNumUnits();
		for i = 0, unitCount - 1 do
			local unit = plot:GetUnit(i);
			if unit ~= nil then
				unit:Kill(true, -1);
			end
		end
	end
}

g_FeaturePlopper =
{
	FeatureType = -1,
	
	Plop =
	function(plot)
		if (g_FeaturePlopper.FeatureType ~= -1) then
			plot:SetFeatureType(g_FeaturePlopper.FeatureType, -1);
		end
	end,
	
	Deplop =
	function(plot)
		plot:SetFeatureType(FeatureTypes.NO_FEATURE, -1);
	end
}

g_RoutePlopper =
{
	RouteType = -1,
	
	Plop =
	function(plot)
		if (g_RoutePlopper.RouteType ~= -1) then
			plot:SetRouteType(g_RoutePlopper.RouteType);
		end
	end,
	
	Deplop =
	function(plot)
		plot:SetRouteType(-1);
	end
}
-- end of VEM, UP updates

g_ResourcePlopper =
{
	ResourceType = -1,
	ResourceAmount = 1,
	
	Plop =
	function(plot)
		if (g_ResourcePlopper.ResourceType ~= -1) then
			plot:SetResourceType(g_ResourcePlopper.ResourceType, g_ResourcePlopper.ResourceAmount);
		end
	end,
	
	Deplop =
	function(plot)
		plot:SetResourceType(-1);
	end
}

g_ImprovementPlopper =
{
	ImprovementType = -1,
	Pillaged = false,
	HalfBuilt = false,
	
	Plop =
	function(plot)
		if (g_ImprovementPlopper.ImprovementType ~= -1) then
			plot:SetImprovementType(g_ImprovementPlopper.ImprovementType);
			if (g_ImprovementPlopper.Pillaged) then
				plot:SetImprovementPillaged(true);
			end
		end
	end,
	
	Deplop =
	function(plot)
		plot:SetImprovementType(-1);
	end
}

g_CityPlopper =
{
	Plop =
	function(plot)
		if (g_PlopperSettings.Player ~= -1) then
			local player = Players[g_PlopperSettings.Player];
			if (player ~= nil) then
				player:InitCity(plot:GetX(), plot:GetY());
			end
		end
	end,
	
	Deplop =
	function(plot)
		local city = plot:GetPlotCity();
		if (city ~= nil) then
			city:Kill();
		end
	end
}

g_PlopperSettings =
{
	Player = -1,
	Plopper = g_UnitPlopper,
	EnabledWhenInTab = false
}

InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_DEBUG][MouseEvents.LButtonUp] = 
function( wParam, lParam )
	local plot = Map.GetPlot( UI.GetMouseOverHex() );
	local plotX = plot:GetX();
	local plotY = plot:GetY();
	local pActivePlayer = Players[Game.GetActivePlayer()];
	local debugItem1 = UI.GetInterfaceModeDebugItemID1();
	
	-- City
	if (UI.debugItem1 == 0) then
		pActivePlayer:InitCity(plotX, plotY);
	-- Unit
	elseif (debugItem1 == 1) then
		local iUnitID = UI.GetInterfaceModeDebugItemID2();
		pActivePlayer:InitUnit(iUnitID, plotX, plotY);
	-- Improvement
	elseif (debugItem1 == 2) then
		local iImprovementID = UI.GetInterfaceModeDebugItemID2();
		plot:SetImprovementType(iImprovementID);
	-- Route
	elseif (debugItem1 == 3) then
		local iRouteID = UI.GetInterfaceModeDebugItemID2();
		plot:SetRouteType(iRouteID);
	-- Feature
	elseif (debugItem1 == 4) then
		local iFeatureID = UI.GetInterfaceModeDebugItemID2();
		plot:SetFeatureType(iFeatureID);
	-- Resource
	elseif (debugItem1 == 5) then
		local iResourceID = UI.GetInterfaceModeDebugItemID2();
		plot:SetResourceType(iResourceID, 5);
	-- Plopper
	elseif (debugItem1 == 6 and
	        type(g_PlopperSettings) == "table" and
	        type(g_PlopperSettings.Plopper) == "table" and
	        type(g_PlopperSettings.Plopper.Plop) == "function") then
	        
		g_PlopperSettings.Plopper.Plop(plot);
		return true; -- Do not allow the interface mode to be set back to INTERFACEMODE_SELECTION
	end
	
	UI.SetInterfaceMode(InterfaceModeTypes.INTERFACEMODE_SELECTION);
	return true;
end

----------------------------------
-- RIGHT MOUSE BUTTON
----------------------------------

InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_DEBUG][MouseEvents.RButtonDown] = 
function( wParam, lParam )
	local plot = Map.GetPlot( UI.GetMouseOverHex() );
	local plotX = plot:GetX();
	local plotY = plot:GetY();
	local debugItem1 = UI.GetInterfaceModeDebugItemID1();
	
	-- Plopper
	if (debugItem1 == 6 and
	        type(g_PlopperSettings) == "table" and
	        type(g_PlopperSettings.Plopper) == "table" and
	        type(g_PlopperSettings.Plopper.Deplop) == "function") then
	        
		g_PlopperSettings.Plopper.Deplop(plot);
	end
	
	return true;
end

InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_DEBUG][MouseEvents.RButtonUp] = 
function( wParam, lParam )
	-- Trap RButtonUp
	return true;
end

InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_RANGE_ATTACK][KeyEvents.KeyDown] = 
function( wParam, lParam )
	if ( wParam == Keys.VK_ESCAPE ) then
		UI.SetInterfaceMode(InterfaceModeTypes.INTERFACEMODE_SELECTION);
		return true;
	elseif (wParam == Keys.VK_NUMPAD1 or wParam == Keys.VK_NUMPAD3 or wParam == Keys.VK_NUMPAD4 or wParam == Keys.VK_NUMPAD6 or wParam == Keys.VK_NUMPAD7 or wParam == Keys.VK_NUMPAD8 ) then
		UI.SetInterfaceMode(InterfaceModeTypes.INTERFACEMODE_SELECTION);
		return DefaultMessageHandler[KeyEvents.KeyDown]( wParam, lParam );
	else
		return DefaultMessageHandler[KeyEvents.KeyDown]( wParam, lParam );
	end
end

InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_CITY_RANGE_ATTACK][KeyEvents.KeyDown] = 
function( wParam, lParam )
	if ( wParam == Keys.VK_ESCAPE ) then
		UI.ClearSelectedCities();
		UI.SetInterfaceMode(InterfaceModeTypes.INTERFACEMODE_SELECTION);
		return true;
	elseif (wParam == Keys.VK_NUMPAD1 or wParam == Keys.VK_NUMPAD3 or wParam == Keys.VK_NUMPAD4 or wParam == Keys.VK_NUMPAD6 or wParam == Keys.VK_NUMPAD7 or wParam == Keys.VK_NUMPAD8 ) then
		UI.ClearSelectedCities();
		UI.SetInterfaceMode(InterfaceModeTypes.INTERFACEMODE_SELECTION);
		return DefaultMessageHandler[KeyEvents.KeyDown]( wParam, lParam );
	else
		return DefaultMessageHandler[KeyEvents.KeyDown]( wParam, lParam );
	end
end

-- this is a default handler for all Interface Modes that correspond to a mission
function missionTypeLButtonUpHandler( wParam, lParam )
	local plot = Map.GetPlot( UI.GetMouseOverHex() );
	if plot then
		local plotX = plot:GetX();
		local plotY = plot:GetY();
		local bShift = UIManager:GetShift();
		local interfaceMode = UI.GetInterfaceMode();
		local eInterfaceModeMission = GameInfoTypes[GameInfo.InterfaceModes[interfaceMode].Mission];
		if eInterfaceModeMission ~= MissionTypes.NO_MISSION then
			if UI.CanDoInterfaceMode(interfaceMode) then

				--XX MOD XX 
				
				local InterfaceControlOverride=false;
				
				dbprint("MISSION ISSUED");				
				local pUnit = UI.GetHeadSelectedUnit();
				if(pUnit~=nil) then
					local unitID = pUnit:GetID();
					
					-- check that unit is not on external control blacklist
					if(blackList[unitID] ~= nil) then
						dbprint("Unit blacklisted. Skipping execution of missionTypeLButtonUpHandler");
						if(blackList[unitID].permanent ~= true) then 
							blackList[unitID] = nil; --remove from blacklist
						end
						return true;
					end
					
					-- make this move control override any units' current orders
					aMissionStack[unitID]=nil;
					
					-- TODO: Move this into separate action handler for REBASE mission
					if(pUnit:GetDomainType() == DomainTypes.DOMAIN_AIR and eInterfaceModeMission == MissionTypes.MISSION_REBASE) then
						
						local range = pUnit:Range() * GameDefines.AIR_UNIT_REBASE_RANGE_MULTIPLIER / 100;
						local thisPlot=pUnit:GetPlot();
						if(thisPlot~=nil and range~=nil and range>0) then
						
							local thisPlotX=thisPlot:GetX();
							local thisPlotY=thisPlot:GetY();
							
							local rebaseToCity=plot:IsCity();
							local rebaseToUnit=nil;
							
							--if rebasing to a ship, we want the ship to restore its alert/sleep/idle state
							if(not rebaseToCity) then
								local unitCount = plot:GetNumUnits();
								for i = 0, unitCount - 1 do
									local pTargetUnit = plot:GetUnit(i);
									if(pTargetUnit~=nil and not pTargetUnit:IsDead() and pTargetUnit:GetDomainType() == DomainTypes.DOMAIN_SEA) then
										--this got to be the target unit, check if it's in the registry
										rebaseToUnit = pTargetUnit;
										local targetUnitID=pTargetUnit:GetID();
										local targetUnitState=pTargetUnit:GetActivityType();
										dbprint("AIR REBASE TO A SHIP. Found target unit, ID:",targetUnitID,"Activity:",targetUnitState);
										
										if(targetUnitState==ActivityTypes.ACTIVITY_SENTRY or targetUnitState==ActivityTypes.ACTIVITY_SLEEP) then
											-- this ship is on alert/sleep. It might have some missions in stack though.
											-- add a mission to restore state as first in stack, with Not Ready Flag set, so event TeleportToHex will only restore the state
											local targetMission=targetUnitState==ActivityTypes.ACTIVITY_SENTRY and GameInfoTypes.MISSION_ALERT or GameInfoTypes.MISSION_SLEEP;
											AddToMissionStack(targetUnitID, "first", targetMission, {nil}, {"NRF"});
										end
									end
								end
							end
							
							-- for air units, make them sleep (or intercept) NEXT turn if they just rebased
							AddToMissionStack(unitID, "only", SleepAlertFortify(pUnit, plot), {nil}, {"NFE", "TF"});
							
							-- functionality for multi-turn routing
							if(range < Map.PlotDistance(thisPlotX, thisPlotY, plotX, plotY)) then
								if(rebaseToCity or (rebaseToUnit~=nil and pUnit:GetSpecialUnitType() == rebaseToUnit:SpecialCargo())) then
									--means unit won't be able to rebase, as it is too far
									--that's a situation in which we could help out!
									local status, coordPath = CalculatePathNodes(thisPlot, plot, range);	
									dbprint("MULTI-NODE REBASING for unit ID:",unitID,"path status:",status,"#turns:",#coordPath);
									
									if(status) then
										InterfaceControlOverride = true;		
										
										--unit should be fortified only if rebasings were successful, hence IFS flag
										AddToMissionStack(unitID, "only", SleepAlertFortify(pUnit, plot), {nil}, {"NFE", "TF", "IFS"});
										
										--a separate mission to rebase between each pair of nodes
										for i,coord in ipairs(coordPath) do
											if(i==1 and rebaseToUnit~=nil) then
												AddToMissionStack(unitID, "first", SeekUnitToRebaseTo, {unitID, rebaseToUnit:GetID()}, {"NFE", "SAFE", {["SMFUN"] = UnitToUnitRebaseTest, ["SMPAR"] = {rebaseToUnit:GetID()}}});
											else
												AddToMissionStack(unitID, "first", GameInfoTypes.MISSION_REBASE, coord, {"NFE", {["SMFUN"] = UnitPositionTest, ["SMPAR"] = coord}});
											end
										end
										
										--this mission will be the very first one, to make sure the move command can be performed with NFE flag
										AddToMissionStack(unitID, "first", GameInfoTypes.MISSION_IDLE, {nil});
										
										ExecuteStack(unitID);

										--show some graphics
										local hexID = ToHexFromGrid( Vector2( plotX, plotY ) );
										Events.SerialEventHexHighlight( hexID, true, 0, "GroupBorder"); 
										UI.SetInterfaceMode(InterfaceModeTypes.INTERFACEMODE_SELECTION);
									end	
								end
							end
							
							--done with all the extra functionality
						end
					end
				end
				
				if(not InterfaceControlOverride) then
					--unit sets orders
					Game.SelectionListGameNetMessage(GameMessageTypes.GAMEMESSAGE_PUSH_MISSION, eInterfaceModeMission, plotX, plotY, 0, false, bShift);
					UI.SetInterfaceMode(InterfaceModeTypes.INTERFACEMODE_SELECTION);
				end
				
				--XX MOD END XX
				
				return true;
			end
		end
	end
	UI.SetInterfaceMode(InterfaceModeTypes.INTERFACEMODE_SELECTION);
	return true;
end

function SeekUnitToRebaseTo(unitID, targetID)
	local player=Players[Game:GetActivePlayer()];
	local pUnit=player:GetUnitByID(unitID);
	local pTarget=player:GetUnitByID(targetID);
	if(pUnit~=nil and pTarget~=nil) then
		local targetX=pTarget:GetX();
		local targetY=pTarget:GetY();
		local targetUnitState=pTarget:GetActivityType();
		--try to rebase if possible
		pUnit:PushMission(GameInfoTypes.MISSION_REBASE, targetX, targetY);
		
		--Restore ship's state if it's on Alert/Sleep
		if(targetUnitState==ActivityTypes.ACTIVITY_SENTRY or targetUnitState==ActivityTypes.ACTIVITY_SLEEP) then
			local targetMission=targetUnitState==ActivityTypes.ACTIVITY_SENTRY and GameInfoTypes.MISSION_ALERT or GameInfoTypes.MISSION_SLEEP;
			-- this ship is on alert/sleep. It might have some missions in stack though.
			-- add a mission to restore state as first in stack, with Not Ready Flag set, so event TeleportToHex will only restore the state
			AddToMissionStack(pTarget:GetID(), "first", targetMission, {nil}, {"NRF"});
		end
		--We modify the stack of different unit than the one currently having its stack executed. So, SeekUnitToRebaseTo is still SAFE
	end
end

function UnitToUnitRebaseTest(pUnit, targetID)
	local player=Players[Game:GetActivePlayer()];
	local pTarget=player:GetUnitByID(targetID);
	
	local test = (pUnit~=nil and pTarget~=nil) and (pUnit:GetX() == pTarget:GetX() and pUnit:GetY() == pTarget:GetY());
	return test;
end


--DEBUG_ITERDATA={};
--DEBUG_CNT=1;
function CalculatePathNodes(plot1, plot2, range)
	local player=Players[Game:GetActivePlayer()];
	
	local x1=plot1:GetX();  local y1=plot1:GetY(); 
	local x2=plot2:GetX();  local y2=plot2:GetY(); 

	local closestU=-1;
	local source=-1;
	local target=-1;
	local fail=true;
	
	local Node={};
	local Prev={};
	local NodeCoordContainer={};
	
	local numCities=player:GetNumCities();
	for i = 0, numCities - 1, 1 do
		local city = player:GetCityByID( i );
		
		ii=i+1;
		Prev[ii]=-1;

		Node[ii]={};				NodeCoordContainer[ii]={};
		Node[ii].x=city:GetX();		NodeCoordContainer[ii].x=Node[ii].x;
		Node[ii].y=city:GetY();		NodeCoordContainer[ii].y=Node[ii].y;
		Node[ii].dist=2000;
		
		if(city==nil) then
			Node[ii]=nil;
		end
		
		if(Node[ii].x==x1 and Node[ii].y==y1) then
			Node[ii].dist=0;	--this is "source", starting point
			source=ii;
		end
		
		if(Node[ii].x==x2 and Node[ii].y==y2) then
			target=ii;			--this is "target", final city
		end
	end
	
	if(source==-1) then --source is not a city. Create a node manually
		source=0;
		Prev[source]=-1;
		
		Node[source]={};		NodeCoordContainer[source]={};
		Node[source].x=x1;		NodeCoordContainer[source].x=x1;
		Node[source].y=y1;		NodeCoordContainer[source].y=y1;
		Node[source].dist=0;
	end
	
	
	if(target==-1) then --target is not a city. Create a node manually
		target=numCities+1;
		Prev[target]=-1;
		
		Node[target]={};		NodeCoordContainer[target]={};
		Node[target].x=x2;		NodeCoordContainer[target].x=x2;
		Node[target].y=y2;		NodeCoordContainer[target].y=y2;
		Node[target].dist=2000;
	end
	
	
		--DEBUG_CNT=1;	---DB	
	closestU=source;	
	while true do
			--DEBUG_ITERDATA[DEBUG_CNT]={};
			--DEBUG_ITERDATA[DEBUG_CNT].Node={};
			--for ee,jj in pairs(Node) do
				--DEBUG_ITERDATA[DEBUG_CNT].Node[ee]=tcopy(jj);
			--end
			--DEBUG_ITERDATA[DEBUG_CNT].closestU=closestU;
			--DEBUG_ITERDATA[DEBUG_CNT].Prev=tcopy(Prev);
			--DEBUG_ITERDATA[DEBUG_CNT].source=source;
			--DEBUG_ITERDATA[DEBUG_CNT].target=target;
	
		dist_u = Node[closestU].dist;
		
			--DEBUG_ITERDATA[DEBUG_CNT].dist_u=dist_u;
	
		if(dist_u > 1000) then break; end;	--no path exists
		
		if(closestU==target) then 	--success, shortest path found!
			fail=false; 
			break; 
		end;
		
		local adjList, closestNode = ListAdjacentNodes(Node, closestU, range);

			--DEBUG_ITERDATA[DEBUG_CNT].adjList=tcopy(adjList);
	
		local newU=closestNode;
		local newUdist=Node[closestNode].dist;
		for i, tmp in pairs(adjList) do
			local alt = Node[closestU].dist + 1;	-- distance to all adjacent nodes is 1 (1 turn)
			if(alt < Node[i].dist) then
				Node[i].dist = alt;
				Prev[i] = closestU;
			end
			if(Node[i].dist < newUdist) then
				newUdist = Node[i].dist;
				newU = i;
			end
		end
			
		Node[closestU]=nil;
		closestU = newU;
	
			--DEBUG_CNT=DEBUG_CNT+1;
	end

	local nodePath={};
	local coordPath={};
	if(not fail) then
		--nodePath is populated in reverse direction, from the target to source (with source missing)
		local i = target;
		local cnt=1;
		while i~=source do
			nodePath[cnt] = i;
			
			local nodeCoord=NodeCoordContainer[i];
			coordPath[cnt] = {nodeCoord.x, nodeCoord.y};
			
			i = Prev[i];
			cnt=cnt+1;
		end
	end		

	return (not fail), coordPath;
end

function ListAdjacentNodes(Node, u, range)
	local adjList={};
	local closestNode=-1;
	local closestDist=4000;
	for i, v in pairs(Node) do
		if(i~=u) then
			local thisVectX=Node[u].x-Node[i].x;
			local thisVectY=Node[u].y-Node[i].y;
			--thisDist=math.sqrt(thisVectX * thisVectX + thisVectY * thisVectY);
			thisDist=Map.PlotDistance(Node[u].x, Node[u].y, Node[i].x, Node[i].y);
			if(thisDist <= range) then
				adjList[i]=true;
			end
			if(Node[i].dist < closestDist) then	--this distance is unrelated to physical hexes distance, it's number of turns to reach the node
				closestDist=Node[i].dist;
				closestNode=i;
			end
		end
	end
	return adjList, closestNode;
end

--Helper function for debugging
function tcopy(t)
  local t2 = {};
  for k,v in pairs(t) do
    t2[k] = v;
  end
  return t2;
end


--this will catch instances of action (e.g. worker build) being issued by pressing a UI button
--those actions are issued in UnitPanel.lua
LuaEvents.UnitActionButtonClicked.Add(
function(action)
	--local interfaceMode = UI.GetInterfaceMode();
	--local eInterfaceModeMission = GameInfoTypes[GameInfo.InterfaceModes[interfaceMode].Mission];
	local missionType=GameInfoActions[action].MissionType;
	local missionData=GameInfoActions[action].MissionData;
	
	dbprint("LuaEvents.UnitActionButtonClicked triggered. Action:",action,"MissionType:",missionType,"MissionData:",missionData);

	if( missionType == MissionTypes.MISSION_BUILD ) then
		
		local pUnit = UI.GetHeadSelectedUnit();
		if(pUnit~=nil) then
			local unitID = pUnit:GetID();
			
			dbprint("LuaEvents.UnitActionButtonClicked: a unit was to told to build. ID:",unitID);
			
			AddToMissionStack(unitID, "only", missionType, {missionData});
			AddRootData(unitID, {["Description"]="unit was told to build something through UI"});
			--no need for ExecuteStack() because a button has just been clicked to start action
		end
	end
	
end);


------------------------------------------------------------------------------------------------------------------------------
--function OnUnitStateChangeDetected(state,unitID,tmp_flag)
--	dbprint("sad");
--	dbprint(state,unitID,tmp_flag);
--end
--Events.UnitActionChanged.Add(OnUnitStateChangeDetected);



InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_MOVE_TO][MouseEvents.LButtonUp] = missionTypeLButtonUpHandler;
InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_MOVE_TO_TYPE][MouseEvents.LButtonUp] = missionTypeLButtonUpHandler;
InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_MOVE_TO_ALL][MouseEvents.LButtonUp] = missionTypeLButtonUpHandler;
InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_ROUTE_TO][MouseEvents.LButtonUp] = missionTypeLButtonUpHandler;
InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_AIRLIFT][MouseEvents.LButtonUp] = missionTypeLButtonUpHandler;
InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_NUKE][MouseEvents.LButtonUp] = missionTypeLButtonUpHandler;
InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_PARADROP][MouseEvents.LButtonUp] = missionTypeLButtonUpHandler;
InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_AIRSTRIKE][MouseEvents.LButtonUp] = missionTypeLButtonUpHandler;
InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_AIR_SWEEP][MouseEvents.LButtonUp] = missionTypeLButtonUpHandler;
InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_REBASE][MouseEvents.LButtonUp] = missionTypeLButtonUpHandler;



InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_AIRSTRIKE][MouseEvents.LButtonUp] = 
function( wParam, lParam )
	local plot = Map.GetPlot( UI.GetMouseOverHex() );
	local plotX = plot:GetX();
	local plotY = plot:GetY();
	local pHeadSelectedUnit = UI.GetHeadSelectedUnit();
	local bShift = UIManager:GetShift();
	local interfaceMode = InterfaceModeTypes.INTERFACEMODE_RANGE_ATTACK;
	
	-- Don't let the user do a ranged attack on a plot that contains some fighting.
	if plot:IsFighting() then
		return true;
	end
		
	-- should handle the case of units bombarding tiles when they are already at war
	if pHeadSelectedUnit and pHeadSelectedUnit:CanRangeStrikeAt(plotX, plotY, true, true) then
		local interfaceMode = UI.GetInterfaceMode();
		local eInterfaceModeMission = GameInfoTypes[GameInfo.InterfaceModes[interfaceMode].Mission];
		Game.SelectionListGameNetMessage(GameMessageTypes.GAMEMESSAGE_PUSH_MISSION, eInterfaceModeMission, plotX, plotY, 0, false, bShift);
    	UI.SetInterfaceMode(InterfaceModeTypes.INTERFACEMODE_SELECTION);
		return true;
	-- Unit Range Strike - special case for declaring war with range strike
	elseif pHeadSelectedUnit and pHeadSelectedUnit:CanRangeStrikeAt(plotX, plotY, false, true) then
		-- Is there someone here that we COULD bombard, perhaps?
		local eRivalTeam = pHeadSelectedUnit:GetDeclareWarRangeStrike(plot);
		if (eRivalTeam ~= -1) then							
			local popupInfo = 
			{
				Type  = ButtonPopupTypes.BUTTONPOPUP_DECLAREWARRANGESTRIKE,
				Data1 = eRivalTeam,
				Data2 = plotX,
				Data3 = plotY
			};
			Events.SerialEventGameMessagePopup(popupInfo);
        	UI.SetInterfaceMode(InterfaceModeTypes.INTERFACEMODE_SELECTION);
			return true;	
		end
	end
						
	return true;
end



InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_RANGE_ATTACK][MouseEvents.LButtonUp] = 
function( wParam, lParam )
	local plot = Map.GetPlot( UI.GetMouseOverHex() );
	local plotX = plot:GetX();
	local plotY = plot:GetY();
	local pHeadSelectedUnit = UI.GetHeadSelectedUnit();
	local bShift = UIManager:GetShift();
	local interfaceMode = InterfaceModeTypes.INTERFACEMODE_RANGE_ATTACK;
	
	-- Don't let the user do a ranged attack on a plot that contains some fighting.
	if plot:IsFighting() then
		return true;
	end
		
	-- should handle the case of units bombarding tiles when they are already at war
	if pHeadSelectedUnit and pHeadSelectedUnit:CanRangeStrikeAt(plotX, plotY, true, true) then
		Game.SelectionListGameNetMessage(GameMessageTypes.GAMEMESSAGE_PUSH_MISSION, MissionTypes.MISSION_RANGE_ATTACK, plotX, plotY, 0, false, bShift);
    	UI.SetInterfaceMode(InterfaceModeTypes.INTERFACEMODE_SELECTION);
		return true;
	-- Unit Range Strike - special case for declaring war with range strike
	elseif pHeadSelectedUnit and pHeadSelectedUnit:CanRangeStrikeAt(plotX, plotY, false, true) then
		-- Is there someone here that we COULD bombard, perhaps?
		local eRivalTeam = pHeadSelectedUnit:GetDeclareWarRangeStrike(plot);
		if (eRivalTeam ~= -1) then							
			local popupInfo = 
			{
				Type  = ButtonPopupTypes.BUTTONPOPUP_DECLAREWARRANGESTRIKE,
				Data1 = eRivalTeam,
				Data2 = plotX,
				Data3 = plotY
			};
			Events.SerialEventGameMessagePopup(popupInfo);
        	UI.SetInterfaceMode(InterfaceModeTypes.INTERFACEMODE_SELECTION);
			return true;	
		end
	end
						
	return true;
end

InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_CITY_RANGE_ATTACK][MouseEvents.LButtonUp] = 
function( wParam, lParam )
	local plot = Map.GetPlot( UI.GetMouseOverHex() );
	local plotX = plot:GetX();
	local plotY = plot:GetY();
	local pHeadSelectedCity = UI.GetHeadSelectedCity();
	local bShift = UIManager:GetShift();
	local interfaceMode = InterfaceModeTypes.INTERFACEMODE_CITY_RANGE_ATTACK;
	
	-- Don't let the user do a ranged attack on a plot that contains some fighting.
	if plot:IsFighting() then
		UI.ClearSelectedCities();
    	UI.SetInterfaceMode(InterfaceModeTypes.INTERFACEMODE_SELECTION);
		return true;
	end
		
	if pHeadSelectedCity and pHeadSelectedCity:CanRangeStrike() then
		if pHeadSelectedCity:CanRangeStrikeAt(plotX,plotY, true, true) then
			Game.SelectedCitiesGameNetMessage(GameMessageTypes.GAMEMESSAGE_DO_TASK, TaskTypes.TASK_RANGED_ATTACK, plotX, plotY);
			local activePlayerID = Game.GetActivePlayer();
			Events.SpecificCityInfoDirty( activePlayerID, pHeadSelectedCity:GetID(), CityUpdateTypes.CITY_UPDATE_TYPE_BANNER);
		end
		UI.ClearSelectedCities();
    	UI.SetInterfaceMode(InterfaceModeTypes.INTERFACEMODE_SELECTION);
		return true;
	end	
	
	return true;
end

-- If the user presses the right mouse button when in city range attack mode, make sure and clear the
-- mode and also clear the selection.
InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_CITY_RANGE_ATTACK][MouseEvents.RButtonUp] = 
function ( wParam, lParam )
	UI.ClearSelectedCities();
   	UI.SetInterfaceMode(InterfaceModeTypes.INTERFACEMODE_SELECTION);
	return true;
end
function EmbarkInputHandler( wParam, lParam )
	local plot = Map.GetPlot( UI.GetMouseOverHex() );
	local plotX = plot:GetX();
	local plotY = plot:GetY();
	local pHeadSelectedUnit = UI.GetHeadSelectedUnit();
	local bShift = UIManager:GetShift();

	if pHeadSelectedUnit then
		if (pHeadSelectedUnit:CanEmbarkOnto(pHeadSelectedUnit:GetPlot(), plot)) then
			Game.SelectionListGameNetMessage(GameMessageTypes.GAMEMESSAGE_PUSH_MISSION, MissionTypes.MISSION_EMBARK, plotX, plotY, 0, false, bShift);
		end
	end
	
	UI.SetInterfaceMode(InterfaceModeTypes.INTERFACEMODE_SELECTION);
end
-- Have either right or left click trigger this
InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_EMBARK][MouseEvents.LButtonUp] = EmbarkInputHandler;
InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_EMBARK][MouseEvents.RButtonUp] = EmbarkInputHandler;

function DisembarkInputHandler( wParam, lParam )
	local plot = Map.GetPlot( UI.GetMouseOverHex() );
	local plotX = plot:GetX();
	local plotY = plot:GetY();
	local pHeadSelectedUnit = UI.GetHeadSelectedUnit();
	local bShift = UIManager:GetShift();

	if pHeadSelectedUnit then
		if (pHeadSelectedUnit:CanDisembarkOnto(plot)) then
			Game.SelectionListGameNetMessage(GameMessageTypes.GAMEMESSAGE_PUSH_MISSION, MissionTypes.MISSION_DISEMBARK, plotX, plotY, 0, false, bShift);
		end
	end
	
	UI.SetInterfaceMode(InterfaceModeTypes.INTERFACEMODE_SELECTION);
end
-- Have either right or left click trigger this
InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_DISEMBARK][MouseEvents.LButtonUp] = DisembarkInputHandler;
InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_DISEMBARK][MouseEvents.RButtonUp] = DisembarkInputHandler;





-- this is a default handler for all Interface Modes that correspond to a mission
--function missionTypeLButtonUpHandler( wParam, lParam )
	--UI.SetInterfaceMode(InterfaceModeTypes.INTERFACEMODE_SELECTION);
	--return true;
--end
--InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_MOVE_TO][MouseEvents.LButtonUp] = missionTypeLButtonUpHandler;
--InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_MOVE_TO_TYPE][MouseEvents.LButtonUp] = missionTypeLButtonUpHandler;
--InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_MOVE_TO_ALL][MouseEvents.LButtonUp] = missionTypeLButtonUpHandler;
--InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_ROUTE_TO][MouseEvents.LButtonUp] = missionTypeLButtonUpHandler;
--InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_AIRLIFT][MouseEvents.LButtonUp] = missionTypeLButtonUpHandler;
--InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_NUKE][MouseEvents.LButtonUp] = missionTypeLButtonUpHandler;
--InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_PARADROP][MouseEvents.LButtonUp] = missionTypeLButtonUpHandler;
----InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_AIRSTRIKE][MouseEvents.LButtonUp] = missionTypeLButtonUpHandler;
--InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_AIR_SWEEP][MouseEvents.LButtonUp] = missionTypeLButtonUpHandler;
--InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_REBASE][MouseEvents.LButtonUp] = missionTypeLButtonUpHandler;
--



InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_SELECTION][MouseEvents.RButtonDown] = 
function( wParam, lParam )
	ShowMovementRangeIndicator();
	UI.SendPathfinderUpdate();
	UpdatePathFromSelectedUnitToMouse();
end

function MovementRButtonUp( wParam, lParam )
	local bShift = UIManager:GetShift();
	local bAlt = UIManager:GetAlt();
	local bCtrl = UIManager:GetControl();
	local plot = Map.GetPlot( UI.GetMouseOverHex() );
	if not plot then
		return true;
	end
	
	local plotX = plot:GetX();
	local plotY = plot:GetY();
	--dbprint("RClick plot x",plotX,"plotY",plotY);
	local pHeadSelectedUnit = UI.GetHeadSelectedUnit();
	UpdatePathFromSelectedUnitToMouse();

	if pHeadSelectedUnit then
	
		-- check that unit is not on external control blacklist
		local unitID = pHeadSelectedUnit:GetID();
		if(blackList[unitID] ~= nil) then
			dbprint("Unit blacklisted. Skipping execution of MovementRButtonUp");
			if(blackList[unitID].permanent ~= true) then 
				blackList[unitID] = nil; --remove from blacklist
			end
			return true;
		end
	
		if (UI.IsCameraMoving() and not Game.GetAllowRClickMovementWhileScrolling()) then
			print("Blocked by moving camera");
			Events.ClearHexHighlights();
			return;
		end
	
		Game.SetEverRightClickMoved(true);
	
		local bBombardEnemy = false;

		-- Is there someone here that we COULD bombard perhaps?
		local eRivalTeam = pHeadSelectedUnit:GetDeclareWarRangeStrike(plot);
		if (eRivalTeam ~= -1) then							
			local popupInfo = 
			{
				Type  = ButtonPopupTypes.BUTTONPOPUP_DECLAREWARRANGESTRIKE,
				Data1 = eRivalTeam,
				Data2 = plotX,
				Data3 = plotY
			};
			Events.SerialEventGameMessagePopup(popupInfo);
			return true;
		end

		-- Visible enemy... bombardment or melee
		if (plot:IsVisibleEnemyUnit(pHeadSelectedUnit:GetOwner()) or plot:IsEnemyCity(pHeadSelectedUnit)) then		
			
			local bNoncombatAllowed = false;
			
			if plot:IsFighting() then
				return true;		-- Already some combat going on in there, just exit
			end
			
			if pHeadSelectedUnit:CanRangeStrikeAt(plotX, plotY, true, bNoncombatAllowed) then
				
				local iMission;
				if (pHeadSelectedUnit:GetDomainType() == DomainTypes.DOMAIN_AIR) then
					iMission = MissionTypes.MISSION_MOVE_TO;		-- Air strikes are moves... yep
				else
					iMission = MissionTypes.MISSION_RANGE_ATTACK;
				end
				
				Game.SelectionListGameNetMessage(GameMessageTypes.GAMEMESSAGE_PUSH_MISSION, iMission, plotX, plotY, 0, false, bShift);
				UI.SetInterfaceMode(InterfaceModeTypes.INTERFACEMODE_SELECTION);
				Events.ClearHexHighlights();
				return true;
			end
			
			--TODO: if unit or city wasn't seen, this won't trigger. But c'mon, does that ever happen?
			if(not pHeadSelectedUnit:IsRanged()) then
				--apparently this unit is not ranged
				--melee attack (by just moving into cell)
				Game.SelectionListGameNetMessage(GameMessageTypes.GAMEMESSAGE_PUSH_MISSION, MissionTypes.MISSION_MOVE_TO, plotX, plotY, 0, false, bShift);
				UI.SetInterfaceMode(InterfaceModeTypes.INTERFACEMODE_SELECTION);
				Events.ClearHexHighlights();
				return true;
			end
			
		end

		-- Garrison in a city	--this functionality will be overriden
		--[[
		local city = plot:GetPlotCity();
		if (city and city:GetOwner() == pHeadSelectedUnit:GetOwner() and pHeadSelectedUnit:IsCanAttack()) then
			local cityOwner = Players[city:GetOwner()];			
			if (not cityOwner:IsMinorCiv() and city:GetGarrisonedUnit() == nil and pHeadSelectedUnit:GetDomainType() == DomainTypes.DOMAIN_LAND) then
				--print("Garrison attempt");
				Game.SelectionListGameNetMessage(GameMessageTypes.GAMEMESSAGE_PUSH_MISSION, MissionTypes.MISSION_GARRISON, plotX, plotY, 0, false, bShift);
				UI.SetInterfaceMode(InterfaceModeTypes.INTERFACEMODE_SELECTION);
				Events.ClearHexHighlights();
				return true;				
			end
		end
		]]--
		
		-- start worker automation
		local city = plot:GetPlotCity();
		if(city and city:GetOwner() == pHeadSelectedUnit:GetOwner() and GameInfo.Units[pHeadSelectedUnit:GetUnitType()].Type=="UNIT_WORKER") then
			
			local keyPressed=bShift or bAlt or bCtrl;
			if(keyPressed) then
				local desiredYieldType=-1; 
				if(bAlt) then desiredYieldType=YieldTypes.YIELD_GOLD;
				elseif(bCtrl) then desiredYieldType=YieldTypes.YIELD_FOOD;
				elseif(bShift) then desiredYieldType=YieldTypes.YIELD_PRODUCTION;
				end
				
				if(g_WrkStandbyFlag==nil) then g_WrkStandbyFlag=false; end;
				
				AutoWorkerIssueOrders(pHeadSelectedUnit:GetID(), plot:GetX(), plot:GetY(), desiredYieldType, g_WrkStandbyFlag);
			else
				--call the popup asking to choose automation options
				tempDataTable.AutoWork={};
				tempDataTable.AutoWork.unitID=pHeadSelectedUnit:GetID();
				tempDataTable.AutoWork.X=plot:GetX();
				tempDataTable.AutoWork.Y=plot:GetY();
				
				LuaEvents.ShowWorkerAutomationPopupMenu();
				
				--the control will return to this point before the user makes a selection
				--another LuaEvent, LuaEvents.ClosedWorkerAutomationPopupMenu will get control once user made a selection
			end
			
			return true;
		end
		
		-- No visible enemy to bombard OR MELEE, just move
		--print("bBombardEnemy check");
		if bBombardEnemy == false then
			local unitID=pHeadSelectedUnit:GetID();
			
			local missionOnArrival=SleepAlertFortify(pHeadSelectedUnit, plot);
			local CFToggle="fortified";
			local moveMission = GameInfoTypes.MISSION_MOVE_TO;
			local isSwappingUnits = false;
		
			if(pHeadSelectedUnit:AtPlot(plot)) then
				--toggle fortification/idle state
				local currentActivity=pHeadSelectedUnit:GetActivityType();
				local currentToggle=(aMissionStack[unitID]~=nil and aMissionStack[unitID].ClickFortifyToggle or nil);
				
				if(currentActivity==ActivityTypes.ACTIVITY_SLEEP or currentActivity==ActivityTypes.ACTIVITY_SENTRY or currentToggle=="fortified") then
					missionOnArrival=GameInfoTypes.MISSION_IDLE;
					CFToggle="not fortified";
				
				elseif(currentActivity == ActivityTypes.ACTIVITY_AWAKE and plot:IsWater() 
						and GameInfo.Units[pHeadSelectedUnit:GetUnitType()].Class == "UNITCLASS_MACHINE_GUN") then
					
					-- deal with a specific case of Machine Gun on board an Amphibious Assault Ship not having an "ALERT" option, only "FORTIFY"
					dbprint("This unit is a machine gun on water tile (AAS?) - use FORTIFY instead of ALERT");
					missionOnArrival = GameInfoTypes.MISSION_FORTIFY;

				end
				
			--elseif(Map.PlotDistance(pHeadSelectedUnit:GetX(), pHeadSelectedUnit:GetY(), plotX, plotY) == 1) then
			elseif(plot:GetNumUnits() > 0) then
				-- swapping 2 neighboring units
				local playerID = Game.GetActivePlayer();
				local unitCount = plot:GetNumUnits();				
				for i = 0, unitCount - 1, 1 do
					local pFoundUnit = plot:GetUnit(i);
					if (pFoundUnit and not pFoundUnit:IsDead() and pFoundUnit:IsCombatUnit() == pHeadSelectedUnit:IsCombatUnit() 
							and pFoundUnit:GetOwner() == playerID and pFoundUnit:GetDomainType() == pHeadSelectedUnit:GetDomainType() 
							and pHeadSelectedUnit:MovesLeft() > 0 and pFoundUnit:MovesLeft() > 0 and pFoundUnit:GetActivityType() == ActivityTypes.ACTIVITY_AWAKE) then  
							
						dbprint("swapping places with another unit");
						-- the 2 units will be swapped
						--moveMission = GameInfoTypes.MISSION_SWAP_UNITS;
						
						-- other unit should be also issued an AlertSleepFortify mission
						local unit2ID = pFoundUnit:GetID();			
						local unit1x = pHeadSelectedUnit:GetX();
						local unit1y = pHeadSelectedUnit:GetY();
						local unit2missionOnArrival = SleepAlertFortify(pHeadSelectedUnit, Map.GetPlotXY(unit1x, unit1y));

						--AddToMissionStack(unit2ID, "only", GameInfoTypes.MISSION_IDLE, {nil}, {{["SMFUN"] = UnitPositionTest, ["SMPAR"] = {unit1x, unit1y}}});
						--AddToMissionStack(unit2ID, "last", unit2missionOnArrival, {nil}, {"TF", "NFE", "IFS"});
						AddToMissionStack(unit2ID, "only", unit2missionOnArrival, {nil}, {"TF"});						
						isSwappingUnits = true;
						
						break;
					end
				end
			end
				
			AddToMissionStack(unitID, "only", moveMission, {plotX, plotY}, {{["SMFUN"] = UnitPositionTest, ["SMPAR"] = {plotX, plotY}}});
			AddToMissionStack(unitID, "last", missionOnArrival, {nil}, {"TF", "NFE", "IFS"});
			AddRootData(unitID, {["Description"]="Unit is RClick-moved", ["ClickFortifyToggle"]=CFToggle});
			
			if(not isSwappingUnits) then
				ExecuteStack(unitID);
			else
				-- swap mission doesn't properly work through PushMission, so we will use interface to execute it
				Game.SelectionListGameNetMessage(GameMessageTypes.GAMEMESSAGE_PUSH_MISSION, MissionTypes.MISSION_MOVE_TO, plotX, plotY, 0, false, bShift);
			end
			
			Events.ClearHexHighlights();
			return true;
		end
	end
end
InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_SELECTION][MouseEvents.RButtonUp] = MovementRButtonUp
InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_MOVE_TO][MouseEvents.RButtonUp] = MovementRButtonUp


--XX MOD XX--

--this is a return point from WorkerAutomatePopup
LuaEvents.ClosedWorkerAutomationPopupMenu.Add(
function(desiredYieldType)
	
	if(desiredYieldType~=nil and g_WrkStandbyFlag~=nil) then
		dbprint("Callback from LuaEvents.ClosedWorkerAutomationPopupMenu",desiredYieldType,g_WrkStandbyFlag);
		AutoWorkerIssueOrders(tempDataTable.AutoWork.unitID, tempDataTable.AutoWork.X, tempDataTable.AutoWork.Y, desiredYieldType, g_WrkStandbyFlag);
	end
	
end);


function CalcTileImproveAttractiveness(tile, isTileWorked, iActivePlayer, playerTechs, desiredYieldType, workerPlot)

	--[[
	local prodBuilds;
	if(desiredYieldType==YieldTypes.YIELD_PRODUCTION) then
		prodBuilds={GameInfoTypes.BUILD_REPAIR,GameInfoTypes.BUILD_SCRUB_FALLOUT,
						GameInfoTypes.BUILD_QUARRY,GameInfoTypes.BUILD_WELL,GameInfoTypes.BUILD_MINE,GameInfoTypes.BUILD_LUMBERMILL,
						GameInfoTypes.BUILD_CAMP,GameInfoTypes.BUILD_PASTURE};
						
	elseif(desiredYieldType==YieldTypes.YIELD_FOOD) then
		prodBuilds={GameInfoTypes.BUILD_REPAIR,GameInfoTypes.BUILD_SCRUB_FALLOUT,
						GameInfoTypes.BUILD_FARM,GameInfoTypes.BUILD_PASTURE,GameInfoTypes.BUILD_PLANTATION};
						
	elseif(desiredYieldType==YieldTypes.YIELD_GOLD) then
		prodBuilds={GameInfoTypes.BUILD_REPAIR,GameInfoTypes.BUILD_SCRUB_FALLOUT,
						GameInfoTypes.BUILD_QUARRY,GameInfoTypes.BUILD_WELL,GameInfoTypes.BUILD_MINE,
						GameInfoTypes.BUILD_CAMP,GameInfoTypes.BUILD_PLANTATION,GameInfoTypes.BUILD_TRADING_POST};
						
	else 	--science
		prodBuilds={GameInfoTypes.BUILD_REPAIR,GameInfoTypes.BUILD_SCRUB_FALLOUT,GameInfoTypes.BUILD_TRADING_POST};
	end
	]]--
	
	local prodBuilds={GameInfoTypes.BUILD_REPAIR,GameInfoTypes.BUILD_SCRUB_FALLOUT,
						GameInfoTypes.BUILD_QUARRY,GameInfoTypes.BUILD_WELL,GameInfoTypes.BUILD_MINE,GameInfoTypes.BUILD_LUMBERMILL,
						GameInfoTypes.BUILD_FARM,GameInfoTypes.BUILD_CAMP,GameInfoTypes.BUILD_PASTURE,
						GameInfoTypes.BUILD_PLANTATION,GameInfoTypes.BUILD_TRADING_POST};
	
	local resourceType=tile:GetResourceType(iActivePlayer);
	local resourceClass=resourceType~=-1 and GameInfo.Resources[resourceType].ResourceClassType or -1;
	local usefulResourceFlag=resourceClass~=-1 and resourceClass~="RESOURCECLASS_BONUS";	
	local featureType=tile:GetFeatureType();
	local forestFlag=featureType~=-1 and (GameInfo.Features[featureType].Type=="FEATURE_FOREST" or GameInfo.Features[featureType].Type=="FEATURE_JUNGLE");
	local falloutFlag=featureType~=-1 and GameInfo.Features[featureType].Type=="FEATURE_FALLOUT";
	local improvementType=tile:GetImprovementType();
	local alreadyImprov=improvementType~=-1;	
	--local tileAlreadyBeingImprovedFlag=IsPlotBeingImproved(tile);
	--local tileWithinOurBordersFlag=tile:GetOwner()==iActivePlayer;
	local hillsPenaltyFlag=(desiredYieldType~=YieldTypes.YIELD_PRODUCTION) and tile:IsHills() or false;
	
	--current attractiveness (improved or unimproved)
	local AttInd = 0;
	for iYield = 0, YieldTypes.NUM_YIELD_TYPES - 1, 1 do
		AttInd = AttInd + tile:CalculateYield(iYield, true);
	end
	
	-- if vanilla civ, the culture needs to be added separately
	local additionalCulture = 0;	
	if(not g_GNK) then
		additionalCulture = tile:GetCulture();
	end
	
	AttInd = (AttInd + additionalCulture + tile:CalculateYield(desiredYieldType, true) * 1.9) / 2.9;	
	
	--find the improvement which maximizes the attractiveness
	local bestImpInd=-100;
	local bestImpID=-1;
	local bestImpSpecBonus=0;
	for i,buildID in pairs(prodBuilds) do
		local worsensFlag=tile:IsBuildRemovesFeature(buildID) or alreadyImprov;
		local scrubFlag=buildID==GameInfoTypes.BUILD_SCRUB_FALLOUT;
		local repairFlag=buildID==GameInfoTypes.BUILD_REPAIR;
		local preventRepairFlag=repairFlag and falloutFlag and alreadyImprov and improvementType==GameInfoTypes.IMPROVEMENT_LUMBERMILL; --handling the case of lumbermills being unrecoverable after a nuke
		local conditionsCheckFlag=(not worsensFlag or usefulResourceFlag or scrubFlag or repairFlag) and not preventRepairFlag;
		--if(tile:CanBuild(buildID) and tileWithinOurBordersFlag and not tileAlreadyBeingImprovedFlag and conditionsCheckFlag) then	--already checked in calling function
		local reqTechID=GameInfoTypes[GameInfo.Builds[buildID].PrereqTech];
		local playerHasPrereqTech=playerTechs:HasTech(reqTechID);
		if(tile:CanBuild(buildID) and conditionsCheckFlag and playerHasPrereqTech) then
		
			local ImpIndRaw = 0;
			for iYield = 0, YieldTypes.NUM_YIELD_TYPES-1, 1 do
				ImpIndRaw = ImpIndRaw + tile:GetYieldWithBuild(buildID, iYield, false, iActivePlayer);
			end
			ImpIndRaw = (ImpIndRaw + (alreadyImprov and 0 or additionalCulture) + tile:GetYieldWithBuild(buildID, desiredYieldType, false, iActivePlayer) * 1.9) / 2.9;
			
			--build-specific bonuses:
			local recoveryBonus=2 + (tile:IsRoute() and 1 or 0);	--bonus for repairing/cleaning fallout
			local repairBonus=(buildID==GameInfoTypes.BUILD_REPAIR) and (recoveryBonus+(usefulResourceFlag and 1 or 0)) or 0;	--when repairing, tiles with strategic/luxury resources get extra bonus so they are reconnected asap
			local falloutScrubBonus=(buildID==GameInfoTypes.BUILD_SCRUB_FALLOUT) and recoveryBonus or 0;
			local chopDownForest=forestFlag and tile:IsBuildRemovesFeature(buildID) and (desiredYieldType==YieldTypes.YIELD_PRODUCTION);	
			
			local ImpSpecBonus = (tile:GetBuildProgress(buildID)>200 and 1 or 0) + repairBonus + falloutScrubBonus + (chopDownForest and 1 or 0);
			
			if(ImpIndRaw + ImpSpecBonus > bestImpInd + (math.random()-0.5)*0.01) then --a tiny random variation is added to have a variety of equally good builds chosen on different occasions
				bestImpInd=ImpIndRaw;
				bestImpID=buildID;
				bestImpSpecBonus=ImpSpecBonus;
			end
		end
	end
	local ImpInd = bestImpInd;

	
	--very rough estimation of number of turns for worker to arrive to this plot, based on physical distance
	local distance=Map.PlotDistance(workerPlot:GetX(), workerPlot:GetY(), tile:GetX(), tile:GetY());
	--distance/1.5 seems a good approximation in terrain without roads and fallout, then add some for roads and hope they go the right way!
	local divisor=1.5 + (workerPlot:IsRoute() and 1.5 or 0) + (tile:IsRoute() and 1.5 or 0) - (falloutScrubFlag and 0.5 or 0);	
	local turns=distance/divisor;	--cap at 5 turns, otherwise improvement attractiveness could become negative if worker starts from afar
	local distanceFactor=math.max(0, 2 - turns/5*2);	--an improvement takes ~5 turns on Standard, and at worst gives +2 something. This is opportunity cost of worker traveling.
	
	
	--apply bonuses
	local tileBonuses = (isTileWorked and 1 or 0);
	local improvementBonuses = (usefulResourceFlag and 2 or 0) + (hillsPenaltyFlag and -0.1 or 0) + distanceFactor + bestImpSpecBonus;
	
	
	--finally, return current attractiveness and improved attractiveness and best build
	local retStruct={};
	retStruct.AttInd=AttInd + tileBonuses;
	retStruct.ImpInd=ImpInd + tileBonuses + improvementBonuses;
	retStruct.MaxAttInd=math.max(retStruct.AttInd,retStruct.ImpInd);
	retStruct.ImpID=bestImpID;
	retStruct.DB_NoBonusImpInd=ImpInd;	
	retStruct.DB_distanceFactor=distanceFactor;
	return retStruct;
	
end

function AutoWorkerIssueOrders(unitID, cityPlotX, cityPlotY, desiredYieldType, standbyFlag)
	dbprint("Call to AutoWorkerIssueOrders", desiredYieldType, standbyFlag);
	
	local iActivePlayer=Game:GetActivePlayer();
	local activeTeam=Teams[Game:GetActiveTeam()];
	local teamTechs=(activeTeam~=nil ) and activeTeam:GetTeamTechs() or nil;
	local pUnit=Players[iActivePlayer]:GetUnitByID(unitID);
	local cityPlot=Map.GetPlot(cityPlotX, cityPlotY);
	local pCity=cityPlot:GetPlotCity();
	if(pUnit~=nil and pCity~=nil) then
	
		local bestWChange=-100;		local bestWTile=nil;	local bestWBuild=-1;
		local worstWImp=100;
		local bestNWImp=-100;		local bestNWTile=nil;	local bestNWBuild=-1;	local bestNWChange=-100;
		
		local initializeW=true;
		local initializeNW=true;
		for i = 0, pCity:GetNumCityPlots() - 1, 1 do

			local tile = pCity:GetCityIndexPlot( i );
			local tileWithinOurBordersFlag=(tile~=nil) and (tile:GetOwner()==iActivePlayer);
			if(tileWithinOurBordersFlag) then

				local tileAlreadyBeingImprovedFlag=IsPlotBeingImproved(tile, pUnit);
				if(not tileAlreadyBeingImprovedFlag and not tile:IsCity()) then

					local isTileWorked= pCity:IsWorkingPlot(tile);	
					attStr=CalcTileImproveAttractiveness(tile,isTileWorked,iActivePlayer,teamTechs,desiredYieldType,pUnit:GetPlot());
					
					--DEBUG message
					local featureType=tile:GetFeatureType();
					dbprint("TILE",tile,"FOOD:",tile:CalculateYield(0, true),"PROD:",tile:CalculateYield(1, true),"GOLD:",tile:CalculateYield(2, true),
									"ATTIND:",string.format("%.4f",attStr.AttInd),"IMPIND:",attStr.ImpInd,"NoBonusIMPIND:",attStr.DB_NoBonusImpInd,"DistFact:",attStr.DB_distanceFactor,"IMP:",attStr.ImpID~=-1 and GameInfo.Builds[attStr.ImpID].Type or "----",
									tile:GetImprovementType()~=-1 and GameInfo.Improvements[tile:GetImprovementType()].Description or "IMPROVEMENT_NONE",
									GameInfo.Terrains[tile:GetTerrainType()].Description,featureType~=-1 and GameInfo.Features[featureType].Description or "FEATURE_NONE");
					--END DEBUG

					--there are 2 competitions: for best improvement among worked tiles and for best improved NW tile against worst improved worked tile
					if(isTileWorked) then
						local AttChange=attStr.ImpInd-attStr.AttInd;
						if(initializeW or (AttChange>bestWChange and AttChange>0)) then --these comparisons are just a tad hacky: I don't explicitly check for whether the tile is already improved (if it is, attStr.ImpInd is a very negative number)
							bestWChange=AttChange;
							bestWTile=tile;
							bestWBuild=attStr.ImpID;
						end
						if(initializeW or (attStr.MaxAttInd<worstWImp and attStr.MaxAttInd>0)) then	
							worstWImp=attStr.MaxAttInd;
						end
						initializeW=false;
					else
						if(initializeNW or (attStr.ImpInd>bestNWImp and attStr.ImpInd>0)) then
							bestNWImp=attStr.ImpInd;
							bestNWTile=tile;
							bestNWBuild=attStr.ImpID;	
							bestNWChange=attStr.ImpInd-attStr.AttInd;
						end
						initializeNW=false;
					end
				end
			end
		end
		
		--and a winner of grand competition among all tiles is...
		dbprint("CRES:", "bestWChange:",bestWChange,"bestNWImp:",bestNWImp,"worstWImp:",worstWImp,"best-worst:",bestNWImp-worstWImp);
		local WorkedTileBetterFlag=(bestWChange>bestNWImp-worstWImp);
		
		local TargetTile=WorkedTileBetterFlag and bestWTile or bestNWTile;
		local TargetBuildID=WorkedTileBetterFlag and bestWBuild or bestNWBuild;
		local TargetAttImprovement=WorkedTileBetterFlag and bestWChange or (bestNWImp-worstWImp);
		
		local NothingToImproveFlag=bestWChange<0 and bestNWChange<0;	--everything around has been improved
		
		dbprint("WIN:","NTIF",NothingToImproveFlag,"WTBF",WorkedTileBetterFlag,"tile",TargetTile,"build",TargetBuildID~=-1 and GameInfo.Builds[TargetBuildID].Type or "----","AttChg",TargetAttImprovement);
		
		Events.ClearHexHighlights();
		CleanupAutoWorkerHighlights();
		local missionReady=false;
		if(not NothingToImproveFlag) then

			--show some graphics
			local hexID = ToHexFromGrid( Vector2( TargetTile:GetX(), TargetTile:GetY() ) );
			local highlightColor = WorkedTileBetterFlag and Vector4( 1, 1, 0, 0.25 ) or Vector4( 0.98, 0.09, 0.8, 1.0 );
			Events.SerialEventHexHighlight( hexID, true, highlightColor, "GroupBorder"); 
			Events.GenericWorldAnchor( GenericWorldAnchorTypes.WORLD_ANCHOR_WORKER, true, TargetTile:GetX(), TargetTile:GetY(), TargetBuildID );
			highlightedTilesList[#highlightedTilesList+1]={["X"]=TargetTile:GetX(), ["Y"]=TargetTile:GetY(), ["Build"]=TargetBuildID};
			
			--issue orders
			AddToMissionStack(unitID, "only", GameInfoTypes.MISSION_MOVE_TO, {TargetTile:GetX(), TargetTile:GetY()}, 
										{{["SMFUN"] = UnitPositionTest, ["SMPAR"] = {TargetTile:GetX(), TargetTile:GetY()}}});
			AddToMissionStack(unitID, "last", GameInfoTypes.MISSION_BUILD, {TargetBuildID}, 
										{"IFS", {["SMFUN"] = ImprovementTest, ["SMPAR"] = {TargetTile:GetX(), TargetTile:GetY(), TargetBuildID}}});
			AddToMissionStack(unitID, "last", AutoWorkerIssueOrders, {unitID, cityPlotX, cityPlotY, desiredYieldType, standbyFlag}, {"IFS", "TF"});		-- Passes its own function handle as argument
			missionReady=true;
		else
			--nothing to do. If on standby, make worker check for improvement opportunities next turn
			if(standbyFlag) then
				AddToMissionStack(unitID, "only", GameInfoTypes.MISSION_SLEEP, {nil}, {"NRF"});	
				AddToMissionStack(unitID, "last", AutoWorkerIssueOrders, {unitID, cityPlotX, cityPlotY, desiredYieldType, standbyFlag}, {"TF"});
				missionReady=true;
			else
				--no orders, let the worker ask for orders
			end
		end
		
		if(missionReady) then
			AddRootData( unitID, {  ["AutoWorkerData"] = {["X"]=TargetTile:GetX(), ["Y"]=TargetTile:GetY()},  ["Description"]="An automated worker"  } );
			ExecuteStack(unitID);
		end
	end
end

function CleanupAutoWorkerHighlights()
	if (highlightedTilesList ~= nil) then
		for i, v in ipairs(highlightedTilesList) do
			Events.GenericWorldAnchor( GenericWorldAnchorTypes.WORLD_ANCHOR_WORKER, false, v.X, v.Y, v.Build );
		end
	end	
	highlightedTilesList = {};
end
Events.UnitSelectionCleared.Add( CleanupAutoWorkerHighlights );


LuaEvents.LTMOD_RemoteControl.Add(
function(unitID, control, place, missionType, missionData, instructionFlags)
	if(unitID~=nil) then
		if(control == "BLACKLIST_NOW") then
			dbprint("");dbprint("LT Remote: unit blacklisted ONCE. ID:",unitID);dbprint("");
			aMissionStack[unitID] = nil;
			blackList[unitID] = {};
			blackList[unitID].permanent = false;
		elseif(control == "BLACKLIST_THIS_TURN") then
			dbprint("");dbprint("LT Remote: unit blacklisted THIS TURN. ID:",unitID);dbprint("");
			aMissionStack[unitID] = nil;
			blackList[unitID] = {};
			blackList[unitID].permanent = true;
			blackList[unitID].oneTurn = true;
		elseif(control == "BLACKLIST_PERMANENT") then
			dbprint("");dbprint("LT Remote: unit blacklisted PERMANENTLY. ID:",unitID);dbprint("");
			aMissionStack[unitID] = nil;
			blackList[unitID] = {};
			blackList[unitID].permanent = true;
		elseif(control == "MISSION") then
			dbprint("");dbprint("LT Remote: unit MISSION assigned. ID:", unitID, place, missionType, missionData, instructionFlags);dbprint("");
			AddToMissionStack(unitID, place, missionType, missionData, instructionFlags);
		end
	end
end);

function AddToMissionStack(unitID, place, missionType, missionData, instructionFlags)
	local stackLength=(aMissionStack[unitID]~=nil and #aMissionStack[unitID] or 0);
	local pos=-1;
	local newf=stackLength==0;
	
	if(place=="only") then
		newf=true;
		pos=1;
	elseif(place=="last") then
		pos=stackLength+1;
	elseif(place=="first") then
		pos=1;
		--shift the existing missions down (if any)
		for i = stackLength, 1, -1 do
			aMissionStack[unitID][i+1]=aMissionStack[unitID][i];
		end
	end
	
	if(pos~=-1) then
		if(newf) then
			aMissionStack[unitID]={};
		end
		aMissionStack[unitID]._update=true;
		aMissionStack[unitID][pos]={};
		aMissionStack[unitID][pos].Type=missionType;
		aMissionStack[unitID][pos].Data=missionData;
		
		if(instructionFlags~=nil) then
			for i,v in ipairs(instructionFlags) do
				if(type(v)=="string") then
					aMissionStack[unitID][pos][v]=true;
				elseif(type(v)=="table") then
					for k,vv in pairs(v) do
						if(type(k)=="string") then
							aMissionStack[unitID][pos][k]=vv;
						end
					end
				end
			end
		end
	else
		dbprint("ERROR @ AddToMissionStack: wrong placement argument");
	end
end

function GetFromMissionStack(unitID, place)

	local missionType=nil;
	local missionData=nil;
			
	local stackLength=(aMissionStack[unitID]~=nil and #aMissionStack[unitID] or 0);	
	if(stackLength>0) then
			local pos=-1;
			if(place=="first") then
				pos=1;
			elseif(place=="last") then
				pos=stackLength;
			end

			if(pos~=-1) then
				missionType=aMissionStack[unitID][pos].Type;
				missionData=aMissionStack[unitID][pos].Data;
			end
	end
	
	return missionType, missionData;
end

function AddRootData(unitID, data)
	if(aMissionStack[unitID]~=nil) then
		for k,v in pairs(data) do
			if(type(k)=="string") then
				aMissionStack[unitID][k]=v;
			else
				dbprint("ERROR @ AddRootData: only string keys are allowed");
			end
		end
	end
end

function ExecuteStack(unitID)		
	dbprint("Call to ExecuteStack. ID:",unitID);
	
	local stackLength=((unitID~=nil and aMissionStack[unitID]~=nil) and #aMissionStack[unitID] or 0);
	if(stackLength>0 and aMissionStack[unitID][1]~=nil) then
		
		local player=Players[Game.GetActivePlayer()];
		local pUnit=player:GetUnitByID(unitID);
		if(pUnit~=nil and not pUnit:IsDead()) then
			
			aMissionStack[unitID]._update=false;
			local missionType=aMissionStack[unitID][1].Type;
			local missionData=aMissionStack[unitID][1].Data;
			
			--retrieve all the instruction flags for this mission
			local NFE = aMissionStack[unitID][1].NFE;			
			local TF = aMissionStack[unitID][1].TF;			
			local RF = aMissionStack[unitID][1].RF;
			local NRF = aMissionStack[unitID][1].NRF;

			local SMFUN = aMissionStack[unitID][1].SMFUN;
			local SMPAR = aMissionStack[unitID][1].SMPAR;
			local IFS = aMissionStack[unitID][1].IFS;
			local IFF = aMissionStack[unitID][1].IFF;
			
			local SAFE = aMissionStack[unitID][1].SAFE;
			
			dbprint("UNIT EXECUTING STACK MISSION. ID: ",unitID,"Unit:",pUnit:GetName(),"Type:",missionType,"Args:",unpack(missionData),
						"NFE:",NFE,"TF:",TF,"RF:",RF,"NRF:",NRF,
						"SMFUN:",SMFUN,"IFS:",IFS,"IFF:",IFF,"_success:",aMissionStack[unitID]._success);
		
			--if forced execution is disabled, make sure unit is idle before executing the mission, otherwise terminate
			if(NFE~=nil and (pUnit:GetActivityType()~=ActivityTypes.ACTIVITY_AWAKE or pUnit:MovesLeft()==0)) then
				return;
			end
			
			--perform a mission unless it has to be skipped based on Success/No Success Flags
			local skip = (IFS ~= nil and aMissionStack[unitID]._success ~= true) or (IFF ~= nil and aMissionStack[unitID]._success ~= false);
			if(not skip) then
				--there are two types of missions: game default and defined here in Lua
				if(type(missionType)=="number") then
					pUnit:PushMission(missionType, unpack(missionData));
				else
					--this mission might modify this unit's stack which we are currently working with
					--so we need to exit ExecuteStack as soon as we get the control back if Update flag has changed
					
					pcall(missionType, unpack(missionData));	
					
					if(SAFE==nil) then
						dbprint("Stopping stack execution after cutom user mission. ID:", unitID);
						return;
					end
				end
			end

			--missions are of two further types: those which end and those which don't (like fortifying)			
			--check Terminal Flag: if set, the mission is issued only once and we don't wait for its completion
			if(TF ~= nil) then 
				aMissionStack[unitID]=nil;
				dbprint("Terminal mission issued. ID:",unitID);
				return;
			end
			
			--check if current mission ended (or if Ready or Not Ready Flag is set) and proceed if so
			local completeFlag=(pUnit:GetActivityType()==ActivityTypes.ACTIVITY_AWAKE and pUnit:MovesLeft()>0) or RF ~= nil or NRF ~= nil;
			if(completeFlag) then
				--if a function is provided to measure the success of the mission, call it and store the result
				if(SMFUN ~= nil and SMPAR ~= nil) then 
					aMissionStack[unitID]._success = SMFUN(pUnit, unpack(SMPAR));
				end
			
				dbprint("Mission completed. Remaining missions: ",stackLength-1);
				--remove this mission and push everything up
				if(stackLength>1) then
					for i = 1, stackLength-1, 1 do
						aMissionStack[unitID][i]=aMissionStack[unitID][i+1];
					end
					aMissionStack[unitID][stackLength]=nil;
					
					--unit is able to undertake the next mission, unless Not Ready Flag is set
					if(NRF == nil) then
						ExecuteStack(unitID);
					end
					return;
				else
					aMissionStack[unitID]=nil;
				end
			end

		else
			--cleanup dead unit
			dbprint("Unit dead, cleaning up. ID:",unitID);
			aMissionStack[unitID]=nil;
		end
	else
		dbprint("stack contents missing");
		aMissionStack[unitID]=nil;
	end
end

function UnitPositionTest(pUnit, x, y)
	local test = pUnit~=nil and (pUnit:GetX() == x and pUnit:GetY() == y);
	dbprint("POSITION TEST:",test);
	return test;
end

function ImprovementTest(pUnit, x, y, buildID)
	
	local plot = Map.GetPlot(x, y);
	local test = pUnit~=nil and (pUnit:GetX() == x and pUnit:GetY() == y) and plot~=nil;
	
	if(test) then
		local improveType = plot:GetImprovementType();
		local improveTest = false;
		
		if(buildID == GameInfoTypes.BUILD_REPAIR) then
			improveTest = not plot:IsImprovementPillaged();
			
		elseif(buildID == GameInfoTypes.BUILD_SCRUB_FALLOUT) then
			local featureType = plot:GetFeatureType();
			improveTest = featureType==-1 or GameInfo.Features[featureType].Type~="FEATURE_FALLOUT";
			
		elseif(improveType ~= -1 and buildID ~= -1) then
			--check if buildID corresponds to the improvement on the tile
			improveTest = GameInfo.Builds[buildID].ImprovementType == GameInfo.Improvements[improveType].Type;	
		end
		
		test = improveTest;
	end
		
	dbprint("IMPROVEMENT TEST:",test);
	return test;
end


function IsPlotBeingImproved(plot, thisWorker)
	local retv=false;
	local unitCount = plot:GetNumUnits();
	for i = 0, unitCount - 1 do
		local punit = plot:GetUnit(i);
		if(punit~=nil and punit~=thisWorker) then 
			local unitType=punit:GetUnitType();
			if(unitType~=-1 and GameInfo.Units[unitType].Type=="UNIT_WORKER") then
				retv=retv or (punit:GetBuildType()~=-1);
			end
		end
	end
	--see if the plot is already allocated to automatic worker
	local player=Players[Game.GetActivePlayer()];
	for unitID, v in pairs(aMissionStack) do
		if( v.AutoWorkerData~=nil and v.AutoWorkerData.X==plot:GetX() and v.AutoWorkerData.Y==plot:GetY() ) then
			local pUnit=player:GetUnitByID(unitID);
			if(pUnit~=nil and not pUnit:IsDead() and pUnit~=thisWorker) then
				retv=true;
			end
		end
	end
	return retv;
end

function UnitMovementFinishes(playerID, unitID)
	dbprint("Events.SerialEventUnitMoveToHexes triggered: called UnitMovementFinishes. ID:",unitID,(aMissionStack[unitID] == nil) and "STACK EMPTY" or "");	
	local player=Players[playerID];
	if(player:IsHuman()) then
		if(aMissionStack[unitID] ~= nil) then
			--The unit might have finished a mission which required it to move. Pass control to stack execution, so next mission could be issued
			ExecuteStack(unitID);
		end
	end
end
Events.SerialEventUnitMoveToHexes.Add( UnitMovementFinishes );

function ExecuteMissionStackOnTurnStart()
		dbprint("");dbprint("");
		dbprint("Events.ActivePlayerTurnStart triggered: called ExecuteMissionStackOnTurnStart. g_LTO_MissionOnStart=",g_LTO_MissionOnStart);	
		local player = Players[Game.GetActivePlayer()];
		
		--if options were changed last turn, now they should take effect
		if(g_LTO_MissionOnStart_set_next_turn ~= nil) then
			g_LTO_MissionOnStart = g_LTO_MissionOnStart_set_next_turn;
			g_LTO_MissionOnStart_set_next_turn = nil;
			dbprint("Set g_LTO_MissionOnStart to ", g_LTO_MissionOnStart);
		end
		
		for unitID, v in pairs(aMissionStack) do
			--check if unit needs to be removed from the LT control blacklist
			if(blackList[unitID]~=nil and blackList[unitID].oneTurn) then
				blackList[unitID] = nil;
			end
			
			--if user wants unit moves to happen before turn end, we need to filter which missions we want to apply now (like fortifying)
			--all others will be postponed (like idling/moving/rebasing/building/all user defined missions)
			local skipExecution=false;
			if(not g_LTO_MissionOnStart) then
				local mtype, mdata = GetFromMissionStack(unitID, "first");
				if(mtype == GameInfoTypes.MISSION_ALERT or mtype == GameInfoTypes.MISSION_FORTIFY or mtype == GameInfoTypes.MISSION_SLEEP 
							or mtype == GameInfoTypes.MISSION_GARRISON or mtype == GameInfoTypes.MISSION_HEAL or mtype == GameInfoTypes.MISSION_SKIP) then
					--these missions will be executed
				else
					--we need to determine current unit state, if unit is idle we will put him on hold for now
					--if not idle, we just don't call ExecuteStack, because the unit won't be asking for orders anyway
					local pUnit=player:GetUnitByID(unitID);
					local unitState=(pUnit~=nil) and pUnit:GetActivityType() or -1;
					if(	unitState == ActivityTypes.ACTIVITY_AWAKE ) then
						AddToMissionStack(unitID, "first", GameInfoTypes.MISSION_IDLE, {nil}, {"RF"});	
						AddToMissionStack(unitID, "first", GameInfoTypes.MISSION_SKIP, {nil}, {"NRF"});						
					else
						skipExecution=true;
					end
					dbprint("ID:",unitID,"Unit:",pUnit:GetName(),"ActivityType:",unitState,"skipExecution:",skipExecution);
				end
			end
			
			if(not skipExecution) then
				ExecuteStack(unitID);		
			end
		end
end
Events.ActivePlayerTurnStart.Add( ExecuteMissionStackOnTurnStart );


function ExecuteMissionStackOnEndTurnDirty()
	local player = Players[Game.GetActivePlayer()];
	if(not g_LTO_MissionOnStart and endTurnExecutedOnTurn~=Game.GetGameTurn() and player:GetEndTurnBlockingType()==-1) then
		dbprint("");dbprint("");
		dbprint("Events.SerialEventEndTurnDirty triggered: called ExecuteMissionStackOnEndTurnDirty. g_LTO_MissionOnStart=",g_LTO_MissionOnStart);	
		
		endTurnExecutedOnTurn=Game.GetGameTurn();	--allow this function to be called only once per turn
		
		for unitID, v in pairs(aMissionStack) do
			ExecuteStack(unitID);
		end
		--local iEndTurnControl = GameInfoTypes.CONTROL_ENDTURN;
		--Game.DoControl(iEndTurnControl)
	end	
end
Events.SerialEventEndTurnDirty.Add( ExecuteMissionStackOnEndTurnDirty );
--Controls.EndTurnButton:RegisterCallback( Mouse.eLClick, ExecuteMissionStackOnEndTurnDirty );


function SleepAlertFortify(pUnit, pPlot)
	--determine what unit combat type this is and decide on appropriate mission
	local unitType = GameInfo.Units[pUnit:GetUnitType()].Type;	
	local combType=pUnit:GetUnitCombatType();
	local domainType=pUnit:GetDomainType();
	local mission;
	local isCity = pPlot:IsCity();
	
	if(domainType == DomainTypes.DOMAIN_LAND) then
		
		if(not g_GNK and isCity and combType~=-1) then
			--any land non-civilian unit can garrison
			-- MISSION_GARRISON seems to be not working in G&K
			mission=GameInfoTypes.MISSION_GARRISON;
		else
			if(combType==GameInfoTypes.UNITCOMBAT_ARCHER) then
				mission=GameInfoTypes.MISSION_ALERT;
			elseif(combType==GameInfoTypes.UNITCOMBAT_ARMOR) then
				mission=GameInfoTypes.MISSION_SLEEP;
			elseif(combType==GameInfoTypes.UNITCOMBAT_GUN) then
				mission=GameInfoTypes.MISSION_FORTIFY;
			elseif(combType==GameInfoTypes.UNITCOMBAT_HELICOPTER) then
				mission=GameInfoTypes.MISSION_SLEEP;
			elseif(combType==GameInfoTypes.UNITCOMBAT_MELEE) then
				mission=GameInfoTypes.MISSION_FORTIFY;
			elseif(combType==GameInfoTypes.UNITCOMBAT_MOUNTED) then
				mission=GameInfoTypes.MISSION_SLEEP;
			elseif(combType==GameInfoTypes.UNITCOMBAT_RECON) then
				mission=GameInfoTypes.MISSION_ALERT;	
			elseif(combType==GameInfoTypes.UNITCOMBAT_SIEGE) then
				mission=GameInfoTypes.MISSION_ALERT;
			else
				--civilians
				if(unitType == "UNIT_GREAT_GENERAL") then
					mission=GameInfoTypes.MISSION_SLEEP;
				else
					mission=GameInfoTypes.MISSION_IDLE;
				end
			end
		end
		
	elseif(domainType == DomainTypes.DOMAIN_AIR) then
		if(pUnit:CurrInterceptionProbability()>0) then
			mission=GameInfoTypes.MISSION_AIRPATROL;
		else
			mission=GameInfoTypes.MISSION_SLEEP;
		end
	elseif(domainType == DomainTypes.DOMAIN_SEA) then
		if(unitType == "UNIT_GREAT_ADMIRAL" or isCity) then
			mission=GameInfoTypes.MISSION_SLEEP;
		else
			mission=GameInfoTypes.MISSION_ALERT;	
		end
	else 
		--what's this unit?
		mission=GameInfoTypes.MISSION_IDLE;
	end

	if(not g_LTO_AlertFortify and mission~=GameInfoTypes.MISSION_GARRISON) then
		--Sleep/Alert/Fortify functionality was disabled in options
		mission=GameInfoTypes.MISSION_IDLE;
	end
	
	return mission;
end

function SetNavalToAlertAfterAirUnitRebasing(tmp1, tmp2, playerID, unitID)
	--NOTE: this function will be called every time after aircraft/missile rebasing
	--but also every time after a loaded ship changes a hex
	--and in several other instances
	
	local player=Players[playerID];
	if(player:IsHuman()) then
		local pu=player:GetUnitByID(unitID);
		if(pu~=nil and not pu:IsDead()) then
			plot=pu:GetPlot();
			
			if(not plot:IsCity()) then
				local unitCount = plot:GetNumUnits();
				for i = 0, unitCount - 1 do
					local pTargetUnit = plot:GetUnit(i);
					if(pTargetUnit~=nil and not pTargetUnit:IsDead() and pTargetUnit:GetDomainType() == DomainTypes.DOMAIN_SEA) then
						--this got to be the target unit, check if it's in the registry
						targetUnitID=pTargetUnit:GetID();
						dbprint("MOD call to SetNavalToAlertAfterAirUnitRebasing",pu:GetName(),unitID,"target:",pTargetUnit:GetName(),targetUnitID);
						-- make unit reset its state (instructions were added at the moment when missile rebased)
						ExecuteStack(targetUnitID);
					end
				end
			end
		end
	end
	
end
Events.SerialEventUnitTeleportedToHex.Add( SetNavalToAlertAfterAirUnitRebasing );


--XX MOD END XX--




function EjectHandler( wParam, lParam )
	local plot = Map.GetPlot( UI.GetMouseOverHex() );
	local plotX = plot:GetX();
	local plotY = plot:GetY();
	--print("INTERFACEMODE_PLACE_UNIT");

	local unit = UI.GetPlaceUnit();
	UI.ClearPlaceUnit();	
	local returnValue = false;
		
	if (unit ~= nil) then
		local unitPlot = unit:GetPlot();
		if (unitPlot ~= nil) then
			local city = unitPlot:GetPlotCity();
			if (city ~= nil) then
				if UI.CanPlaceUnitAt(unit, unitPlot, plot) then
					--Network.SendCityEjectGarrisonChoice(city:GetID(), plotX, plotY);
					returnValue =  true;					
				end
			end
		end
	end
	
	UI.SetInterfaceMode(InterfaceModeTypes.INTERFACEMODE_SELECTION);
	
	return returnValue;
end
InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_PLACE_UNIT][MouseEvents.LButtonUp] = EjectHandler;
InterfaceModeMessageHandler[InterfaceModeTypes.INTERFACEMODE_PLACE_UNIT][MouseEvents.RButtonUp] = EjectHandler;


DefaultMessageHandler[MouseEvents.RButtonUp] =
function( wParam, lParam )
	UI.SetInterfaceMode(InterfaceModeTypes.INTERFACEMODE_SELECTION);
end


local mouseX, mouseY;
	
----------------------------------------------------------------
-- Input handling 
----------------------------------------------------------------
function InputHandler( uiMsg, wParam, lParam )

		-- update global var rButtonDown
		if uiMsg == MouseEvents.RButtonDown then
			rButtonDown = true;
		elseif uiMsg == MouseEvents.RButtonUp then
			rButtonDown = false;
		end
	
		local interfaceMode = UI.GetInterfaceMode();
		local currentInterfaceModeHandler = InterfaceModeMessageHandler[interfaceMode];
		if currentInterfaceModeHandler and currentInterfaceModeHandler[uiMsg] then
			return currentInterfaceModeHandler[uiMsg]( wParam, lParam );
		elseif DefaultMessageHandler[uiMsg] then
			return DefaultMessageHandler[uiMsg]( wParam, lParam );
		end
	
		return false;
end
ContextPtr:SetInputHandler( InputHandler );

---------------------------------------------------------------- 
-- Deal with a new path from the path finder 
-- this is a place holder implementation to give an example of how to handle it       
----------------------------------------------------------------
function OnUIPathFinderUpdate( thePath )
	--g_pathFromSelectedUnitToMouse = thePath;
	UpdatePathFromSelectedUnitToMouse();
end
Events.UIPathFinderUpdate.Add( OnUIPathFinderUpdate );

function OnMouseMoveHex( hexX, hexY )
	--local pPlot = Map.GetPlot( hexX, hexY);
	local interfaceMode = UI.GetInterfaceMode();	
	if (interfaceMode == InterfaceModeTypes.INTERFACEMODE_SELECTION and rButtonDown) or (interfaceMode == InterfaceModeTypes.INTERFACEMODE_MOVE_TO) then
		UI.SendPathfinderUpdate();
	end
	mouseX = hexX;
	mouseY = hexY;
end
Events.SerialEventMouseOverHex.Add( OnMouseMoveHex );

function OnClearUnitMoveHexRange()
	Events.ClearHexHighlightStyle(pathBorderStyle);
	Events.ClearHexHighlightStyle(attackPathBorderStyle);
end
Events.ClearUnitMoveHexRange.Add( OnClearUnitMoveHexRange );

function OnStartUnitMoveHexRange()
	Events.ClearHexHighlightStyle(pathBorderStyle);
	Events.ClearHexHighlightStyle(attackPathBorderStyle);
end
Events.StartUnitMoveHexRange.Add( OnStartUnitMoveHexRange );

function OnAddUnitMoveHexRangeHex(i, j, k, attackMove, unitID)
	if attackMove then
		Events.SerialEventHexHighlight( Vector2( i, j ), true, turn1Color, attackPathBorderStyle );
	else
		Events.SerialEventHexHighlight( Vector2( i, j ), true, turn1Color, pathBorderStyle );
	end
end
Events.AddUnitMoveHexRangeHex.Add( OnAddUnitMoveHexRangeHex );

--Events.EndUnitMoveHexRange.Add( OnEndUnitMoveHexRange );

local g_PerPlayerStrategicViewSettings = {}
----------------------------------------------------------------
-- 'Active' (local human) player has changed
----------------------------------------------------------------
function OnActivePlayerChanged(iActivePlayer, iPrevActivePlayer)
	
	if (iPrevActivePlayer ~= -1) then
		g_PerPlayerStrategicViewSettings[ iPrevActivePlayer + 1 ] = InStrategicView();
	end
	
	if (iActivePlayer ~= -1 ) then
		if (g_PerPlayerStrategicViewSettings[ iActivePlayer + 1] and not InStrategicView()) then
			ToggleStrategicView();
		else
			if (not g_PerPlayerStrategicViewSettings[ iActivePlayer + 1] and InStrategicView()) then
				ToggleStrategicView();
			end
		end
	end
end
Events.GameplaySetActivePlayer.Add(OnActivePlayerChanged);